"""
    Copyright 2015-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import eggwrapper as _eggwrapper
from .. import _interface
from .. import _site

class EHH(object):
    """
    Extended Haplotype Homozygosity statistics.
    Statistics can be computed for unphased genotypic data.

    The usage of this class is to: to first set the core haplotypes by
    passing a :class:`.Site` to :meth:`.set_core` and then to
    load repetitively distant sites (always with increasing distance
    from the core), through :meth:`.load_distant`, until the list of
    sites to process is exhausted or one of the thresholds has been
    reached.

    In order to process distant sites using the same core region to the
    opposite direction, or to use a different core region, it is always
    required to call :meth:`!set_core` again (this is the only way to
    reset the instance).

    After at least one distant site site is loaded, EHH statistics can
    be accessed using there accessors. EHH statistics fell into four
    categories:

    1. Raw EHH statistics, provided for each loaded distant sites and
       separately for each core haplotype. See the methods:
       :meth:`.get_EHH`, :meth:`.get_EHHc`, and :meth:`.get_rEHH`.

       .. admonition:: Reference

           Sabeti P.C., D.E. Reich, J.M. Higgins, H.Z.P. Levine,
           D.J. Richter, S.F. Schaffner, S.B. Gabriel, J.V. Platko, N.J.
           Patterson, G.J. McDonald, H.C. Ackerman, S.J. Campbell, D.
           Altshuler, R. Cooper, D. Kwiatkowski, R. Ward & E.S. Lander.
           2002. Detecting recent positive selection in the human genome
           from haplotype structure. *Nature*\ \xa0\ **419**\ : 832-837.

    2. Integrated EHH statistics, computed separately for each core
       haplotype and incremented at each provided sites until the EHH
       value reaches the thresholds provided as the *EHH_thr* and
       *EHHc_thr* arguments to :meth:`!set_core`. See the methods
       :meth:`.get_iHH`, :meth:`.get_iHHc`, and :meth:`.get_iHS`. The
       methods :meth:`.done_EHH` and :meth:`.done_EHHc` allow to check
       whether the threshold has been reached for all genotypes.

       .. admonition:: Reference

           Voight B.F., S. Kudaravalli, X. Wen & J.K. Pritchard.
           2006. A map of recent positive selection in the human genome.
           *PLoS Biol*\ \xa0\ **4**\ : e772.

    3. Whole-site EHHS statistic and its integrated version iES, which
       is incremented as long that the EHHS value is larger than or equal to
       the threshold provided as the *EHHS_thr* argument to
       :meth:`!set_core`. See the methods :meth:`.get_EHHS` and
       :meth:`.get_iES`. If data are unphased, a specific EHHS estimate
       based on homozygosity is provided (EHHG, with its iEG integrated
       counter-part).

       .. admonition:: Reference

           Tang K., K.R. Thornton & M. Stoneking. 2007. A new
           approach for using genome scans to detect recent positive
           selection in the human genome. *PLoS Biol.*\ \xa0\ **5**\ : e171.

    4. EHH, EHHS, and EHHG decay statistics, that give the minimal distance at
       which, respectively, EHH starts to be smaller than the threshold
       provided as the *EHH_thr* argument to :meth:`!set_core`, EHHS
       starts to be smaller than the threshold provided as the
       *EHHS_thr* argument to :meth:`!set_core`, and EHHG
       starts to be smaller than the threshold provided as the
       *EHHG_thr* argument to :meth:`!set_core`. EHH decay is computed
       separately for each core haplotype. See the methods
       :meth:`.get_dEHH`, :meth:`.get_dEHHS`, and :meth:`.get_dEHHG`. The values are not
       available until the respective threshold is reached (``None`` is
       returned). The method  :meth:`.done_EHH` allows to check whether
       the threshold for the EHH decay has been reached for all core
       haplotypes. The maximum and average value of the EHH decay across
       all core haplotypes can be accessed with :meth:`.get_dEHH_max`
       and :meth:`.get_dEHH_mean`, respectively.

       .. admonition:: Reference

           Ram\u00EDrez-Soriano A., S.E. Ramos-Onsins, J. Rozas,
           F. Calafell & A. Navarro. 2008. Statistical power analysis of
           neutrality tests under demographic expansions, contractions and
           bottlenecks with recombination. *Genetics*\ \xa0\ **179** \:
           555-567.

    In all cases,
    ``None`` is returned when the value is not available (no data
    loaded, division by zero in the case of ratios, or threshold not
    reached in the case of decay statistics).

    The thresholds must all be within the range between 0 and 1.
    """

    def __init__(self):
        self._obj = _eggwrapper.EHH()
        self._struct = _eggwrapper.StructureHolder()
        self._valid = False
        self._distance = 0.0
        self._K_core = None
        self._K_tot = None
        self._K_cur = None
        self._R = None

    def set_core(self, site, struct=None, min_freq=None, min_sam=2,
                    EHH_thr=None, EHHc_thr=None, EHHS_thr=None, EHHG_thr=None, crop_EHHS=False):
        """
        Specify the core site. If the
        instance already contains data, it will be reset.

        :param site: a :class:`.Site` instance. The site must have
            a specified :py:obj:`~.Site.position`. It is
            assumed that data are phased (samples should be phased; but
            it is possible to load genotypes whose phase is unknown, as
            long as the genotypes are phased at the individual level:
            see *indiv_struct*).

        :param struct: This argument must be set to a non-``None``
            value if genotypic data are provided and if unphased versions
            of statistics should be computed. If  not ``None``, this argument
            should either be a :class:`.Structure` object mapping samples
            to individuals (other structure levels are not considered),
            or an integer specifying the ploidy (in the latter case, samples
            of any individuals are supposed to be consecutive in the
            site. It is still required that individuals are phased (that
            is, that they are entered in the same order in all distant sites).
            The integer 0 is accepted as a synonym of ``None`` (data will be
            treated as haploid).

        :param min_freq: minimal absolute frequency for haplotypes
            (haplotypes with lower frequencies are ignored). By default,
            all haplotypes are considered.

        :param min_sam: minimal absolute number of non-missing samples to
            compute statistics or stop incrementing them (for statistics
            expressed per core haplotype, number of samples for this
            haplotype).

        :param EHH_thr: threshold determining when iHH should stop
            integrating and dEHH must be evaluated. By default (``None``), iHH is
            permanently incremented and dEHH is not evaluated at all.

        :param EHHc_thr: threshold determining when iHH should stop
            integrating and dEHH must be evaluated. By default (if ``None``), use the
            same value as for *EHH_thr*.

        :param EHHS_thr: threshold determining when iES should stop
            integrating and dEHHS must be evaluated. By default (``None``), iES is
            permanently incremented and dEHHS is not evaluated at all.

        :param EHHG_thr: threshold determining when iEG (iES computed for
            unphased data) should stop
            integrating and dEHHG (equivalent to dEHHS)
            must be evaluated. Must be ``None`` if
            *genotypes* is ``True``. By default (``None``), iEG is
            permanently incremented and dEHHG is not evaluated at all.
            By default: equal to EHHS_thr.

        :param crop_EHHS: if True, set values of EHHS that are below the
            threshold to 0 to emulate the behaviour of the R package
            rehh (also affects iES).
        """

        # check validity of instance
        if not isinstance(site, _site.Site): raise TypeError('a Site instance is expected')
        if min_freq is None: min_freq = 1
        elif min_freq < 1: raise ValueError('minimal frequency must be strictly larger than 0')
        if min_sam < 2: raise ValueError('minimal number of non-missing samples must be at least 2')
        if site.position is None: raise ValueError('core site must have a position')

        # load core site and get core statistics
        if EHH_thr is None: EHH_thr = 0.0
        elif EHH_thr < 0.0 or EHH_thr > 1.0: raise ValueError('invalid value for `EHH_thr`')
        if EHHc_thr is None: EHHc_thr = EHH_thr
        elif EHHc_thr < 0.0 or EHHc_thr > 1.0: raise ValueError('invalid value for `EHHc_thr`')
        if EHHS_thr is None: EHHS_thr = 0.0
        elif EHHS_thr < 0.0 or EHHS_thr > 1.0: raise ValueError('invalid value for `EHHS_thr`')
        if EHHG_thr is None: EHHG_thr = EHHS_thr
        elif EHHG_thr < 0.0 or EHHG_thr > 1.0: raise ValueError('invalid value for `EHHG_thr`')

        if struct == 0:
            struct = None
        elif struct is None:
            pass
        elif isinstance(struct, _interface.Structure):
            struct = struct._obj
            if struct.get_req() > site._obj.get_ns(): raise ValueError('structure does not match site')
        elif isinstance(struct, int):
            ns = site.ns
            pl = struct
            if ns % pl != 0: raise ValueError('number of samples is not a multiple of the specified ploidy')
            self._struct.mk_dummy_structure(ns//pl, pl)
            struct = self._struct
        else:
            raise TypeError('invalid type for `struct` argument')
        self._obj.set_core(site._obj, struct, EHH_thr, EHHc_thr, EHHS_thr, EHHG_thr, min_freq, min_sam, crop_EHHS)

        self._K_core = self._obj.K_core()
        if self._K_core == 0: raise ValueError('cannot analyse core site: no valid data')

        # initialize flags
        self._valid = True
        self._ns = site._obj.get_ns()

    def load_distant(self, site):
        """
        Process a distant site. The core site must have been specified.
        The site must have a specified :py:obj:`~.Site.position`.

        :param site: a :class:`.Site` instance containing data
            for the distant site. It must be consistent with the core
            site (same list of samples, and in the same order). Missing
            data are supported.
        """

        # check validity of argument
        if not self._valid: raise ValueError('no core site has been loaded')
        if site._obj.get_ns() != self._ns: raise ValueError('distant site must be consistent with core')
        if site.position is None: raise ValueError('distant site must have a position')

        # pass site
        self._obj.load_distant(site._obj)

    @property
    def num_haplotypes(self):
        """
        Number of core haplotypes. ``None`` if
        core has not been set.
        """
        return self._K_core

    @property
    def cur_haplotypes(self):
        """
        Current number of haplotypes. ``None`` if core has not been set,
        equal to :obj:`.num_haplotypes` if no distant site has been
        loaded.
        """
        return self._obj.K_cur()

    @property
    def nsam(self):
        """ Current number of non-missing samples. """
        if not self._valid: raise ValueError('cannot access statistics (no core site provided)')
        return self._obj.num_avail_tot()

    def nsam_core(self, hap):
        """ Number of samples for a core haplotype.
            Number of non-missing samples for one of the core haplotypes. """
        if not self._valid: raise ValueError('no data processed')
        if hap < 0 or hap >= self._K_core: raise IndexError('invalid haplotype index')
        return self._obj.num_avail_core(hap)

    def nsam_hap(self, hap):
        """ Number of samples for a haplotype.
            Number of non-missing samples for one of the current haplotypes. """
        if not self._valid: raise ValueError('no data processed')
        if hap < 0 or hap >= self._obj.K_cur(): raise IndexError('invalid haplotype index')
        return self._obj.num_avail_cur(hap)

    def get_EHH(self, i):
        """ Get an EHH value.
        Get the EHH value for the last processed distant site for core
        haplotype *i*. Return ``None`` if the value cannot be computed (no
        available samples).
        """
        if not self._valid: raise ValueError('no data processed')
        if i<0 or i>=self._K_core: raise IndexError('invalid haplotype index')
        value = self._obj.EHHi(i)
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_EHHc(self, i):
        """ Get an EHHc value.
        Get the EHHc value for the last processed distant site for core
        haplotype *i*. Return ``None`` if the value cannot be computed (no
        available samples).
        """
        if not self._valid: raise ValueError('no data processed')
        if i<0 or i>=self._K_core: raise IndexError('invalid haplotype index')
        value = self._obj.EHHc(i)
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_rEHH(self, i):
        """ Get a rEHH value.
        Get the rEHH value for the last processed distant site for core
        haplotype *i*. Return ``None`` if the ratio cannot be computed
        (no available sample or division by zero).
        """
        if not self._valid: raise ValueError('no data processed')
        if i<0 or i>=self._K_core: raise IndexError('invalid haplotype index')
        value = self._obj.rEHH(i)
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_iHH(self, i):
        """ Get an iHH value.
        Get the iHH value for the last processed distant site for core
        haplotype *i* . Return ``None`` if the value cannot be computed (no
        available samples).
        """
        if not self._valid: raise ValueError('no data processed')
        if i<0 or i>=self._K_core: raise IndexError('invalid haplotype index')
        value = self._obj.iHH(i)
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_iHHc(self, i):
        """ Get an iHHc value.
        Get the iHHc value for the last processed distant site for core
        haplotype *i*. Return ``None`` if the value cannot be computed (no
        available samples).
        """
        if not self._valid: raise ValueError('no data processed')
        if i<0 or i>=self._K_core: raise IndexError('invalid haplotype index')
        value = self._obj.iHHc(i)
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_iHS(self, i):
        """ Get an iHS value.
        Get the iHS value for the last processed distant site for core
        haplotype *i*. Return ``None`` if the ratio cannot be computed
        (no available sample or division by zero).
        """
        if not self._valid: raise ValueError('no data processed')
        if i<0 or i>=self._K_core: raise IndexError('invalid haplotype index')
        value = self._obj.iHS(i)
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def done_EHH(self):
        """Completed EHH calculation.
        Return ``True`` if the values of iHH for all core haplotypes
        have finished integrating (and all dEHH values have been
        evaluated).
        """
        if not self._valid: raise ValueError('no data processed')
        return self._obj.num_EHH_done() == self._K_core

    def done_EHHc(self):
        """ Completed EHHc calculation.
        Return ``True`` if the values of iHHc for all core haplotypes
        have finished integrating (and all dEHH values have been
        evaluated).
        """
        if not self._valid: raise ValueError('no data processed')
        return self._obj.num_EHHc_done() == self._K_core

    def get_EHHS(self):
        """ Get EHHS value.
        Get the EHHS value for the last processed distant site. Return ``None`` if
        the value cannot be computed (no available samples).
        """
        if not self._valid: raise ValueError('no data processed')
        value = self._obj.EHHS()
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_EHHG(self):
        """ Get EHHG value.
        Get the EHHS (computed with genotypes) value for the last processed distant site. Return ``None`` if
        the value cannot be computed (no available samples, no indiv_struct option used).
        """
        if not self._valid: raise ValueError('no data processed')
        value = self._obj.EHHG()
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_iES(self):
        """ Get iES value.
        Get the iES value for the last processed distant site. Return ``None`` if
        the value cannot be computed (no available samples).
        """
        if not self._valid: raise ValueError('no data processed')
        value = self._obj.iES()
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_iEG(self):
        """ Get iEG value.
        Get the iES (computed with genotypes) value for the last processed distant site. Return ``None`` if
        the value cannot be computed (no available samples or indiv_struct option was not used).
        """
        if not self._valid: raise ValueError('no data processed')
        value = self._obj.iEG()
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_dEHH(self, i):
        """ EHH decay.
        Get the EHH decay distance for core haplotype *i*. Return ``None`` if
        the EHH threshold has not been reached.
        """
        if not self._valid: raise ValueError('no data processed')
        if i<0 or i>=self._K_core: raise IndexError('invalid haplotype index')
        value = self._obj.dEHH(i)
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_dEHHc(self, i):
        """ EHHc decay.
        Get the EHHc decay distance for core haplotype *i*. Return ``None`` if
        the EHHc threshold has not been reached.
        """
        if not self._valid: raise ValueError('no data processed')
        if i<0 or i>=self._K_core: raise IndexError('invalid haplotype index')
        value = self._obj.dEHHc(i)
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_dEHH_max(self):
        """ Maximum EHH decay.
        Get the maximum EHH decay distance across core haplotypes. Return ``None`` if
        the EHH threshold has not been reached for at least one of the
        core haplotypes.
        """
        if not self._valid: raise ValueError('no data processed')
        value = self._obj.dEHH_max()
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_dEHH_mean(self):
        """ Average EHH decay.
        Get the average EHH decay distance across core haplotypes. Return ``None`` if
        the EHH threshold has not been reached for at least one of the
        core haplotypes.
        """
        if not self._valid: raise ValueError('no data processed')
        value = self._obj.dEHH_mean()
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_dEHHS(self):
        """ EHHS decay.
        Get the EHHS decay distance. Return ``None`` if the EHHS threshold has not
        been reached.
        """
        if not self._valid: raise ValueError('no data processed')
        value = self._obj.dEHHS()
        if value == _eggwrapper.UNDEF: return None
        else: return value

    def get_dEHHG(self):
        """ EHHG decay.
        Get the EHHS (computed with genotypes) decay distance. Return ``None`` if the EHHG threshold has not
        been reached.
        """
        if not self._valid: raise ValueError('no data processed')
        value = self._obj.dEHHG()
        if value == _eggwrapper.UNDEF: return None
        else: return value



"""
    Copyright 2015-2021 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import eggwrapper as _eggwrapper
from .. import _interface, alphabets
from ..tools import _code_tools
from .. import _site

class CodingDiversity(object):
    """
    Detection of synonymous and non-synonymous sites.
    This class processes alignments with a reading frame specification
    in order to detect synonymous and non-synonymous variable positions.
    It provides basic statistics, but it can also filter data to let the
    user compute all other statistics on synonymous-only, or
    non-synonymous-only variation (e.g. :math:`\pi` or ``D``).

    The constructor takes optional arguments. By default, build an empty
    instance. If arguments are passed, they must match the signature of
    :meth:`~.CodingDiversity.process` that will be called.

    Note that codons with missing data are never considered, even if the
    resolution of all possibilities consistent translate to the same
    nucleotide. This applies to stop and start codons. Alternative start
    codons are not considered either.
    """

    def _get_site(self):
        if len(self._site_pool) > 0: return self._site_pool.pop(0)
        else: return _site.Site(alphabets.codons)

    def _defaults(self):
        self._num_tot = 0
        self._num_eff = 0
        self._num_NS = 0.0
        self._num_S = 0.0
        self._num_pol_single = 0
        self._num_multiple_hits = 0
        self._num_multiple_alleles = 0
        self._num_pol_NS = 0
        self._num_pol_S = 0
        self._num_stop = 0

    def __init__(self, *args, **kwargs):
        # internal processing helpers
        self._site_pool = []
        self._sites_S = []
        self._sites_NS = []
        self._alleles_S = []
        self._alleles_NS = []
        self._positions_S = []
        self._positions_NS = []

        # run process
        if len(args) + len(kwargs) > 0: self.process(*args, **kwargs)

        # set internal variables to default values
        else: self._defaults()

    def process(self, align, code=1, 
            struct=None, max_missing=0.0,
            skipstop=True, raise_stop=False,
            multiple_alleles=False, multiple_hits=False):
        """
        Process an alignment. It this instance already had data in
        memory, they will all be erased.

        :param align: an :class:`.Align` instance containing the coding
            sequence to process. The alphabet must be codons.

        :param code: genetic code identifier (see :ref:`here <genetic-codes>`).
            Required to be an integer among the valid values. The
            default value is the standard genetic code.

        :param struct: a class:`Structure` object specifying samples to
            include and outgroup samples to skip for analysis (outgroup
            samples are ignored to determine if site is polymorphic,
            coding, or non-coding, but included in generated sites).
            By default, use all samples.

        :param max_missing: maximum relative proportion of missing data (per
            codon site) to allow (including stop codons if *skipstop* if
            ``true``). By default, all codon sites with any missing data
            are excluded. Outgroup is not considered.

            .. note::
                Here, *max_missing* is a relative proportion to allow
                using the same value for different alignments that might
                not have the same number of samples (to avoid reevaluating
                *max_missing* if the user wants the same maximum rate
                of missing data). In other functions, *max_missing* is
                the maximum *number* of missing and is an integer.

        :param skipstop: if ``True``, stop codons are treated as missing
            data and skipped. If so, potential mutations to stop codons
            are not taken into account when estimating the number of
            non-synonymous sites. Warning (this may be
            counter-intuitive): it actually assumes that stop codons are
            not biologically plausible and considers them as missing
            data. On the other hand, if *skipstop* is ``False``, it
            considers stop codons as if they were valid amino acids.
            This option has no effect if *raise_stop* is ``True``.

        :param raise_stop: raise a :exc:`ValueError` if a
            stop codon is met. If ``True``, *skipstop* has no effect.
            Outgroup is not considered.

        :param multiple_alleles: include coding sites with more than
            two alleles (regardless of whether mutations hit the same
            position within the triplet). If there are more than two
            different codons, they must either encode for the same amino acid
            or all encode for different amino acids (otherwise the site is
            rejected). If more than one of the three codon position are
            mutated, the option *multiple_hits* is considered.

        :param multiple_hits: include coding sites for which more than
           one of the three positions has changed (regardless of the
           number of alleles). If there are more than two alleles, the
           option *multiple_alleles* is also considered.
        """
        if not isinstance(align, _interface.Align): raise TypeError('an Align instance is required')
        if align._alphabet._obj.get_type() != 'codons': raise ValueError('alphabet must be codons')
        if code not in _code_tools._codes: raise ValueError('unknown genetic code: {0}'.format(code))
        code = _code_tools._codes[code]
        if struct is None:
            struct = _eggwrapper.StructureHolder()
            struct.mk_dummy_structure(align.ns, 1)
        else: struct = struct._obj

        # if raise_stop is on: force skipstop to be false to make them appear
        if raise_stop and skipstop: skipstop = False

        # return all CodingSite's to stock
        self._site_pool.extend(self._sites_S)
        self._site_pool.extend(self._sites_NS)
        del self._sites_S[:]
        del self._sites_NS[:]
        del self._alleles_S[:]
        del self._alleles_NS[:]
        del self._positions_S[:]
        del self._positions_NS[:]

        # get the first coding site
        current = self._get_site()

        # initialize variables
        self._num_tot = align.ls
        self._num_eff = 0
        self._num_stop = 0
        self._num_NS = 0.0
        self._num_S = 0.0
        self._num_pol_single = 0
        self._num_multiple_hits = 0
        self._num_multiple_alleles = 0
        self._num_pol_NS = 0
        self._num_pol_S = 0
        ns = struct.get_ni()
        max_missing = int(max_missing * ns)

        # process all codon sites
        for idx in range(self._num_tot):
            # process site (don't care about missing data)
            current._obj.reset()
            current._obj.process_align(align._obj, idx, struct)
                # WE ASSUME HERE THAT INGROUP SAMPLES ARE PACKED AT THE FRONT OF THE SITE

            # count the number of stop codon and missing data
            nstop = 0
            nmiss = 0
            for i in range(ns):
                if current._obj.get_sample(i) < 0: nmiss += 1
                elif code.is_stop_unsmart(current._obj.get_sample(i)):
                    if raise_stop: raise ValueError('stop codon found in sequences')
                    nstop += 1
            if skipstop: nmiss += nstop

            # check stop codon
            if nstop > 0: self._num_stop += 1 # incremented even if skipstop is False

            # skip if too many missing data (but still count stop codon)
            if nmiss > max_missing: continue
            good = True

            # check number of alleles (exclude if no alleles)
            alleles = set()
            for i in range(ns):
                a = current._obj.get_sample(i)
                if a >= 0 and (not skipstop or not code.is_stop_unsmart(a)):
                    alleles.add(a)
            na = len(alleles)
            if na < 1:
                continue

            if na == 2:
                self._num_pol_single += 1 # decremented below if the two alleles differ at more than one position

            if na > 2:
                self._num_multiple_alleles += 1
                good &= multiple_alleles

            # check for multiple hits
            na1, na2, na3 = map(len, map(set, zip(* map(alphabets.codons.get_value, alleles))))
            if (na1 > 1) + (na2 > 1) + (na3 > 1) > 1:
                self._num_multiple_hits += 1
                good &= multiple_hits
                if na == 2: self._num_pol_single -= 1

            # exclude site if too multiple alleles/hits
            if not good: continue

            # process alleles
            if na > 1:
                aas = set(map(code.translate, alleles))
                if len(aas) == 1: SYN = True
                elif len(aas) == na: SYN = False
                elif len(aas) < na: continue # skip because ambiguous syn/non-syn
                else: raise RuntimeError('unexpected error in CodingDiversity')

                if SYN:
                    self._num_pol_S += 1
                    self._sites_S.append(current)
                    self._alleles_S.append(alleles)
                    self._positions_S.append(idx)
                else:
                    self._num_pol_NS += 1
                    self._sites_NS.append(current)
                    self._alleles_NS.append(alleles)
                    self._positions_NS.append(idx)

            # count site as exploitable
            self._num_eff += 1
            num_NS = 0.0
            num_S = 0.0
            c = 0
            for i in range(ns):
                codon = current._obj.get_sample(i)
                if codon >= 0 and (not skipstop or not code.is_stop_unsmart(current._obj.get_sample(i))):
                    num_NS += code.NSsites(codon, skipstop)
                    num_S += code.Ssites(codon, skipstop)
                    c += 1
            if c > 0:
                self._num_NS += num_NS / c
                self._num_S += num_S / c

            if na > 1:
                # generate a new `current` site
                current = self._get_site()

    @property
    def num_codons_tot(self):
        """
        Total number of considered codon sites. Only
        complete codons have been considered, but this value includes
        codons that have been rejected because of missing data.
        """
        return self._num_tot

    @property
    def num_codons_eff(self):
        """
        Number of analysed codon sites. Like
        :attr:`.num_codons_tot` but excluding sites rejected because of
        missing data.
        """
        return self._num_eff

    @property
    def num_codons_stop(self):
        """
        Number of codon sites with at least one codon stop.
        """
        return self._num_stop

    @property
    def num_sites_NS(self):
        """
        Estimated number of non-synonymous sites. Note that the total
        number of sites per codon is always 3.

        The numbers of non-synonymous and synonymous sites are estimated
        using the method of Nei & Gojobori (*Mol. Biol. Evol.* 1986
        **3**:418-426).
        """
        return self._num_NS

    @property
    def num_sites_S(self):
        """
        Estimated number of synonymous sites. Note that the total number
        of sites per codon is always 3.
        """
        return self._num_S

    @property
    def num_pol_single(self):
        """
        Number of polymorphic coding sites with only one mutation. All these
        sites are always included.
        """
        return self._num_pol_single

    @property
    def num_multiple_alleles(self):
        """
        Number of polymorphic coding sites with more than two alleles. These
        sites are included only if *multiple_alleles* is ``True`` except those
        who mix synonymous and non-synonymous changes (they can be rejected
        if there are more than two alleles in total as well).
        """
        return self._num_multiple_alleles

    @property
    def num_multiple_hits(self):
        """
        Number of polymorphic codons for which more than one position is
        changed. These sites are included only if *multiple_hits* is ``True``
        and depdenting on the total number of alleles.
        """
        return self._num_multiple_hits

    @property
    def num_pol_NS(self):
        """
        Number of polymorphic codon sites with only non-synonymous variation.
        """
        return self._num_pol_NS

    @property
    def num_pol_S(self):
        """
        Number of polymorphic codon sites with only synonymous variation.
        """
        return self._num_pol_S

    @property
    def sites_S(self):
        """
        List of sites with only synonymous variation
        """
        return self._sites_S

    @property
    def positions_S(self):
        """
        List of positions of sites with only synonymous variation
        """
        return [site.position for site in self._sites_S]

    @property
    def sites_NS(self):
        """
        List of sites with only non-synonymous variation
        """
        return self._sites_NS

    @property
    def positions_NS(self):
        """
        List of positions of sites with only non-synonymous variation
        """
        return [site.position for site in self._sites_NS]



"""
    Copyright 2015-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import eggwrapper as _eggwrapper
from .. import _interface
from .. import _site

def _get_stats(pld, multiple_policy, min_freq):
    if pld.num_alleles1() > 2 or pld.num_alleles2() > 2:
        if multiple_policy == 'forbid':
            raise ValueError('at least one locus has more than two alleles and this is not allowed')
        elif multiple_policy == 'main':
            freq1 = [pld.freq1(i) for i in range(pld.num_alleles1())]
            freq2 = [pld.freq2(i) for i in range(pld.num_alleles2())]
            a1 = [freq1.index(max(freq1))]
            a2 = [freq2.index(max(freq2))]
        elif multiple_policy == 'average':
            a1 = [i for i in range(pld.num_alleles1()) if pld.freq1(i) >= min_freq]
            a2 = [i for i in range(pld.num_alleles2()) if pld.freq2(i) >= min_freq]
        else:
            raise ValueError('invalid value for `multiple_policy` option: {0}'.format(multiple_policy))
    else:
        a1 = [0]
        a2 = [0]

    n = 0
    D = 0.0
    Dp = 0.0
    r = 0.0
    rsq = 0.0
    for i in a1:
        for j in a2:
            n += 1
            pld.compute(i, j)
            D += pld.D()
            Dp += pld.Dp()
            r += pld.r()
            rsq += pld.rsq()
    if n == 0:
        return {'D': None, 'Dp': None, 'r': None, 'rsq': None}
    if n > 1:
        D /= n
        Dp /= n
        r /= n
        rsq /= n
    return {'D': D, 'Dp': Dp, 'r': r, 'rsq': rsq, 'n': n}

def pairwise_LD(locus1, locus2, struct=None, multiple_policy='main', min_freq=0):
    """
    Linkage disequilibium between a pair of loci.
    
    :class:`.Site` instances must be used to describe loci. Only the order of
    samples is considered and both loci must have the same number of
    samples.

    :param locus1: A :class:`.Site` instance.

    :param locus2: A :class:`.Site` instance.

    :param struct: A :class:`.Structure` object used to describe the
        individual level. If specified, genotypes are automatically
        described and genotypic linkage is processed. By default, consider
        all samples as originating from haploid individuals.

    :param multiple_policy: Determine what is done if either input locus
        has more than two alleles. Possible values are, ``"forbid"``
        (raise an exception if this occurs), ``"main"`` (take
        the most frequent allele of each locus) and ``"average"`` (
        compute the unweighted average over all possible pair of
        alleles). More options might be added in future versions. This
        option is ignored if both loci have less than three alleles. If
        ``"main"`` and there are several equally most frequent alleles,
        the first-occurring one is used (arbitrarily).

    :param min_freq: Only used if at least one site has more than two
        alleles and *multiple_policy* is set to *average*. Set the
        minimum absolute frequency to consider an allele.

    :return: A :class:`dict` of linkage disequilibrium statistics. In case
        statistics cannot be computed (either site fixed, or less than
        two samples with non-missing data at both sites), computed
        values are replaced by ``None``. ``n`` gives the number of pairs
        of alleles considered (0 if statistics are not computed at all).
    """

    if locus1.ns != locus2.ns: raise ValueError('the number of samples must match in the two loci')
    if locus1.ns < 2: raise ValueError('the number of samples must be at least 2')

    if struct is None:
        struct = _eggwrapper.StructureHolder()
        struct.mk_dummy_structure(locus1.ns, 1)
        struct2 = struct
    else:
        if struct.req_ns > locus1.ns: raise ValueError('mismatch with structure')
        struct2 = struct.make_auxiliary()._obj
        struct = struct._obj

    site1 = _eggwrapper.Genotypes()
    site2 = _eggwrapper.Genotypes()
    site1.process(locus1._obj, struct)
    site2.process(locus2._obj, struct)
    frq1 = _eggwrapper.FreqHolder()
    frq1.setup_structure(struct2)
    frq1.process_site(site1.site())
    frq2 = _eggwrapper.FreqHolder()
    frq2.setup_structure(struct2)
    frq2.process_site(site2.site())
    ld = _eggwrapper.PairwiseLD()
    if not ld.process(site1.site(), site2.site(), frq1, frq2, struct2, 0, 1.0):
        return {'D': None, 'Dp': None, 'r': None, 'rsq': None, 'n': 0}
    return _get_stats(ld, multiple_policy, min_freq)

def matrix_LD(align, stats, struct=None, multiple_policy='main',
        min_freq=0, min_n=2, max_maj=1.0, positions=None):
    """
    Linkage disequilibrium statistics between all
    pairs of sites. The computed statistics
    are selected by an argument of this function. Return a matrix (as a
    nested :class:`list`) of the requested statistics. In all cases, all
    pairs of sites are present in the returned matrices. If statistics
    cannot be computed, they are replaced by :data:`None`.

    The available statistics are:

    * ``d`` -- Distance between sites of the pairs.
    * ``D`` -- Standard linkage disequilibrium.
    * ``Dp`` -- Lewontin's D'.
    * ``r`` -- Correlation coefficient.
    * ``rsq`` -- Equivalent to r\ :sup:`2`.

    :param align: An :class:`.Align` instance.

    :param stats: Requested statistic or statistics (see list of
        available statistics above, as a single string or as a list of
        one or more of these statistics (in any order).

    :param struct: A :class:`.Structure` object used to describe the
        individual level. If specified, genotypes are automatically
        described and genotypic linkage is processed. By default, consider
        all samples as originating from haploid individuals.

    :param multiple_policy: Specify what is done for pairs of sites for
        which at least one locus has only one allele. See
        :func:`pairwise_LD` for further description.

    :param min_freq: Only used if at least one site has more than two
        alleles and depending on the value of *multiple_policy*.  See
        :func:`pairwise_LD` for further description.

    :param min_n: Minimum number of samples used (this value must
        always be larger than 1). Sites not fulfilling this criterion
        will be dropped.

    :param max_maj: Maximum relative frequency of the majority allele.
        Sites not fulfilling this criterion will be dropped.

    :param positions: A sequence of positions, whose length must match
        the number of sites of the provided alignment. Used in the
        return value to describe the used sites, and, if requested, to
        compute the distance between sites. By default, the position of
        sites in the original alignment is used.

    :return: A :class:`tuple` with two items: first is the :class:`list` of
        positions of sites used in the matrix (a subset of the sites of
        the provided alignment), with positions provided by the
        corresponding argument (by default, the index of sites); second
        is the matrix, as the nested lower half matrix. The matrix
        contains items for all i and j indexes with 0 <= j <= i < n
        where n is the number of retained sites. The content of the
        matrix is represented by a single value (if a single statistic
        has been requested) or as a list of 1 or more values (if a list
        of 1 or more, accordingly, statistics have been requested), or
        ``None`` for the diagonal or if the pairwise comparison was
        dropped for any reason.
    """

    # initialize local variables
    mLD = _eggwrapper.MatrixLD()
    retained = []
    if struct is None:
        struct = _eggwrapper.StructureHolder()
        struct.mk_dummy_structure(align.ns, 1)
    else:
        struct = struct._obj
    mLD.set_structure(struct)

    # check arguments
    min_n = int(min_n)
    if min_n < 2: raise ValueError('too small value for `min_n` argument: {0}'.format(min_n))
    max_maj = float(max_maj)
    if max_maj < 0.5 or max_maj > 1.0: raise ValueError('invalid value for `max_maj` argument: {0}'.format(max_maj))

    if positions is None:
        positions = list(map(float, range(align.ls)))
    elif len(positions) != (struct.get_ni() + struct.get_no()):
        raise ValueError('list of positions does not have the right number of items')

    if stats in ['D', 'Dp', 'r', 'rsq']:
        multi = False
        stats = [stats]
    else:
        multi = True

    # make a list of variable sites
    genosite = _eggwrapper.Genotypes()
    genosites = [] # need to be kept to perform calculations
    site = _site.Site()
    final_positions = []
    sd = _eggwrapper.SiteDiversity()
    frq = _eggwrapper.FreqHolder()
    frq.setup_structure(struct)
    for i in range(align.ls):
        site.from_align(align, i)
        genosite.process(site._obj, struct)
        frq.process_site(genosite.site())
        if (sd.process(frq)&2) != 0 and sd.Aing() > 1.0:
            final_positions.append(positions[i])
            mLD.load(genosite, positions[i])
            genosites.append(genosite)
            genosite = _eggwrapper.Genotypes()

    matrix = [[None for j in range(i+1)] for i in range(len(genosites))]

    # compute LD
    mLD.computeLD(min_n, max_maj)

    # extract requested values
    for i in range(mLD.num_pairs()):
        pld = mLD.pairLD(i)
        idx2 = mLD.index1(i)
        idx1 = mLD.index2(i) # reverse indexes
        computed_stats = _get_stats(pld, multiple_policy, min_freq)
        matrix[idx1][idx2] = []
        for stat in stats:
            if stat == 'd': matrix[idx1][idx2].append( final_positions[idx1] - final_positions[idx2] )
            elif stat == 'D': matrix[idx1][idx2].append( computed_stats['D'] )
            elif stat == 'Dp': matrix[idx1][idx2].append( computed_stats['Dp'] )
            elif stat == 'r': matrix[idx1][idx2].append( computed_stats['r'] )
            elif stat == 'rsq': matrix[idx1][idx2].append( computed_stats['rsq'] )
            else: raise ValueError('invalid statistic code: `{0}`'.format(stat))
        if not multi:
            matrix[idx1][idx2] = matrix[idx1][idx2][0]

    # return
    return final_positions, matrix



import pathlib, re, glob, os
import setuptools, sysconfig
from setuptools.command.build_ext import build_ext

##### GET VERSION ######################################################

with pathlib.Path('src', 'egglib', '__init__.py').open() as f:
    if mo := re.search('__version__ *= *[\'"]((\d+\.)\d\.\d+(?:(?:a|b|rc)\d*)?)[\'"]', f.read()):
        version = mo.group(1)
    else:
        raise RuntimeError("Unable to find version string.")

##### OVERRIDE BUILD_EXT COMMAND #######################################

class build_ext2(build_ext):
    def run(self):
        # get the build path (there should be a better way to get it!)
        build_path = str(pathlib.Path(self.get_ext_fullpath('libegglib')).parent / 'egglib')
        self.library_dirs.append(build_path)

        # get egglib installation options (at some point, we should be able to use the --config-settings option)
        DEBUG = os.environ.get('DEBUG', '0')
        if DEBUG == '0':
            for ext in self.extensions: ext.extra_compile_args.extend(['-g', '-O3'])
        elif DEBUG == '1':
            for ext in self.extensions: ext.extra_compile_args.extend(['-g', '-O0'])
            for ext in self.extensions: ext.define_macros.append(('DEBUG', 1))
        else: raise ValueError(f'invalid value for DEBUG: {DEBUG} (accepted values are 0 and 1)')

        # idem for the options --no-htslib and --require-htslib
        HTSLIB = os.environ.get('HTSLIB', '2')
        if HTSLIB == '0':
            assert self.extensions[-1].name == 'egglib.io._vcfparser'
            del self.extensions[-1]
            super().run() # no error should occur
        elif HTSLIB == '1':
            super().run() # an error might occur but the user should be prepared
        elif HTSLIB == '2':
            try:
                super().run()
            except setuptools.errors.CompileError as e:
                print(e)
                assert self.extensions[-1].name == 'egglib.io._vcfparser'
                del self.extensions[-1]
                return
        else:
            raise ValueError(f'invalid value for HTSLIB: {HTSLIB} (accepted values are 0, 1 and 2)')

##### EXTENSION MODULES ################################################

lib = 'egglib' + str(pathlib.Path(sysconfig.get_config_var('EXT_SUFFIX')).stem)
cpath = pathlib.Path('src', 'cfiles')
cpppath = pathlib.Path('src', 'cppfiles')

def filelist(path, names, ext):
    return [path.joinpath(i).with_suffix(ext) for i in libfiles]

libegglib = setuptools.Extension('egglib.libegglib',
                    sources=[str(cpath.joinpath(n).with_suffix('.c')) for n in ['random']],
                    language='c')

random = setuptools.Extension('egglib.random',
                    sources=[str(cpath.joinpath('randomwrapper').with_suffix('.c'))],
                    language='c',
                    libraries = [lib],
                    extra_link_args = ["-Wl,-rpath=$ORIGIN/."])

vcf = setuptools.Extension('egglib.io._vcfparser',
                    sources=[str(cpath.joinpath('vcfwrapper').with_suffix('.c'))],
                    language='c',
                    libraries = [lib, 'hts'],
                    extra_link_args = ["-Wl,-rpath=$ORIGIN/."])

binding = setuptools.Extension('egglib._eggwrapper',
                    sources=glob.glob(str(pathlib.Path(cpppath, '*.cpp'))),
                    language='c++',
                    swig_opts=['-python', '-c++', '-builtin', '-Wall'],
                    include_dirs = [os.path.join('src', 'cfiles')],
                    libraries = [lib],
                    extra_link_args = ["-Wl,-rpath=$ORIGIN/."])

extensions = [libegglib, binding, random, vcf]

##### MAIN PACKAGE #####################################################

pkg_list = ['egglib',
            'egglib.cli',
            'egglib.test',
            'egglib.test.data',
            'egglib.test.base',
            'egglib.test.coalesce',
            'egglib.test.stats',
            'egglib.test.tools',
            'egglib.test.io',
            'egglib.test.wrappers',
            'egglib.coalesce',
            'egglib.io',
            'egglib.stats',
            'egglib.tools',
            'egglib.wrappers']

setuptools.setup(
    cmdclass={'build_ext': build_ext2},
    version=version,
    package_dir={'egglib': os.path.join(r'src', 'egglib')},
    packages=pkg_list,
    package_data={
        'egglib.wrappers': ['apps.conf'],
        'egglib.test.data': ['*.fas', '*.fa', '*.gb', '*.txt',
                             '*.gff3', '*.sta', '*.gnl', '*.fg',
                             '*.aln', '*.vcf', '*.vcfi', '*.bed',
                             '*.hap', '*.inp', '*.gpop', '*.cds',
                             '*.asnb', '*.bcf', '*.clu']},
    ext_modules=extensions
)



"""
    Copyright 2015-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

        EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import re, operator, os, sys
from io import StringIO
from .. import eggwrapper as _eggwrapper
from .. import alphabets, _interface

_SOFA = {
    'SO:0000704': 'gene',
    'SO:0000234': 'mRNA',
    'SO:0000147': 'exon',
    'SO:0000316': 'cds',
    'SO:0000188': 'intron',
    'SO:0000610': 'polyA_sequence',
    'SO:0000553': 'polyA_site',
    'SO:0000204': 'five_prime_UTR',
    'SO:0000205': 'three_prime_UTR'
}

class GFF3(object):
    """
    Import GFF3 genome annotation data.
    Read General Feature Format (GFF)-formatted (version 3) genome
    annotation data from a file specified by name or from a provided
    string.

    See the description of the GFF3 format at `<http://www.sequenceontology.org/gff3.shtml>`_.

    All features are loaded into memory and can be processed interactively.

    :param source: name of a GFF3-formatted file. To pass a string,
        use the factory method :meth:`from_string`.
    :param strict: by default, apply strictly GFF3 requirements. See
        below.

    .. _gff3-strict:

    The ``strict`` argument can be set to ``False`` to support CDS
    features lacking a phase.

    Some observations regarding this implementation:

    * The validity of parent-child relationship with respect to the
      Sequence Ontology is not enforced, meaning that this parser
      allows any type of feature to be parent to any feature of any
      type.

    * The Sequence Ontology accession number of types of features related
      to genes are automatically mapped to the type name (such as
      "SO:0000704", which is translated as "gene").

    * Discontinuous features are required to have matching attributes
      (except start, stop, and phase for CDS features) but subsequent
      segments are allowed to skip attributes if the first
      segment defines them.

    * fasta-formatted sequences are imported but are currently not
      checked for consistency with annotations and defined sequence
      regions.

    :class:`.io.GFF3` instances are iterable. The expression:

    .. code-block:: python

        >>> for feat in GFF3:
        ...     ...

    is equivalent to:

    .. code-block:: python

        >>> for feat in GFF3.iter_features():
        ...     ...
    """

    @classmethod
    def from_string(cls, string, strict=True):
        """
        Import data from a GFF3-formatted string. This is a class method,
        which can be called as shown below to create a new object:

        .. code-block:: python

            >>> gff3 = egglib.io.GFF3.from_string(gff3_string)

        :param string: GFF3-formatted string.
        :param strict: apply strict GFF3 specifications (see :ref:`here <gff3-strict>`).
        :return: A new :class:`.io.GFF3` instance.

        """
        obj = cls.__new__(cls)
        obj._strict = strict
        obj._parse(StringIO(string))
        return obj

    def __init__(self, fname, strict=True):
        self._strict = strict
        f = open(fname, mode='r')
        self._parse(f)

    def _parse(self, f):  # read a file or file-like object
        # initialize parameters
        self._regions = {}
        self._feature_ontology = []
        self._attribute_ontology = []
        self._source_ontology = []
        self._species = None
        self._genome_build = None
        self._sequences = None
        self._seqid = []
        self._mapping = {}
        self._num_tot = 0
        self._num_top = 0
        self._IDs = set() # IDs of all features
        self._open_features = {} # only those with an ID

        # check first line
        mo = re.match('##gff-version[ \t]+(3(\.\d+(\.\d+)?)?)$', f.readline().strip())
        if mo is None: raise IOError('invalid first line of GFF3 file (invalid gff-version directive or invalid GFF version)')
        self._version = mo.group(1)

        # read all lines
        self._lineno = 1 # offset because of header line
        self._f = f
        self._pos = self._f.tell()
        while True:
            self._line = f.readline()
            if self._line == '': break
            self._lineno += 1
            self._readline(self._line)
        self._close_objects()

        # sort features within seqid entries
        for key in self._mapping:
            self._mapping[key].sort(key=lambda feat: (feat.start, feat.end))
            for feat in self._mapping[key]:
                feat._descendants.sort(key=operator.attrgetter('start'))
                feat._all_parts.sort(key=operator.attrgetter('start'))

    def _readline(self, line): # process a line
        if line.rstrip() == '###': self._close_objects()
        elif line[0] == '>': self._get_sequences()
        elif line.rstrip() == '##FASTA':
            self._pos = self._f.tell()
            self._get_sequences()
        elif line[:2] == '##': self._get_directive()
        elif line[0] == '#': pass
        else: self._get_annotation()
        self._pos = self._f.tell()

    def _close_objects(self): # close all open features
        for ID, feat in self._open_features.items():
            feat._close()
        self._open_features.clear()

    def _process_text(self, text): # expand escape expressions
        for hit in set(re.findall('(%[0-9A-Fa-f]{2})', text)):
            text = text.replace(hit, chr(int('0x'+hit[1:], 16)))
        return text

    def _get_directive(self): # import a directive
        args = self._line.split()
        if args[0] == '##sequence-region':
            if len(args) != 4: raise IOError('invalid sequence-region directive [line: {0}]'.format(self._lineno))
            seqid, start, end = args[1:]
            seqid = self._process_text(seqid)
            try:
                start = int(start) - 1
                end = int(end) - 1
            except ValueError: raise IOError('invalid sequence-region directive [line: {0}]'.format(self._lineno))
            if start < 0 or end < start: raise IOError('invalid sequence-region directive [line: {0}]'.format(self._lineno))
            if seqid in self._regions: raise IOError('invalid sequence-region directive (seqid already defined) [line: {0}]'.format(self._lineno))
            self._regions[seqid] = start, end
        elif args[0] == '##feature-ontology':
            if len(args) != 2: raise IOError('invalid feature-ontology directive [line: {0}]'.format(self._lineno))
            self._feature_ontology.append(self._process_text(args[1]))
        elif args[0] == '##attribute-ontology':
            if len(args) != 2: raise IOError('invalid attribute-ontology directive [line: {0}]'.format(self._lineno))
            self._attribute_ontology.append(self._process_text(args[1]))
        elif args[0] == '##source-ontology':
            if len(args) != 2: raise IOError('invalid source-ontology directive [line: {0}]'.format(self._lineno))
            self._source_ontology.append(self._process_text(args[1]))
        elif args[0] == '##species':
            if self._species is not None: raise IOError('species directive repeateError(d [line: {0}]'.format(self._lineno))
            self._species = ' '.join(map(self._process_text, args[1:]))
        elif args[0] == '##genome-build':
            if len(args) != 3: raise IOError('invalid genome-build directive [line: {0}]'.format(self._lineno))
            if self._genome_build is not None: raise IOError('genome-build directive repeated [line: {0}]'.format(self._lineno))
            self._genome_build = tuple(map(self._process_text, args[1:]))
        else:
            raise IOError('invalid directive: {0} [line: {1}]'.format(args[0], self._lineno))

    def _get_annotation(self): # import an annotation feature

        # separate the 9 columns
        bits = self._line.rstrip().split('\t')
        if len(bits) != 9: raise IOError('invalid annotation line [line: {0}]'.format(self._lineno))

        # seqid
        if bits[0] == '.': raise IOError('invalid annotation line (seqid cannot be missing) [line: {0}]'.format(self._lineno))
        else: seqid = self._process_text(bits[0])

        # source
        if bits[1] == '.': source = None
        else: source = self._process_text(bits[1])

        # type
        if bits[2] == '.': raise IOError('invalid annotation line: type must be defined [line: {0}]'.format(self._lineno))
        type_ = self._process_text(bits[2])
        type_ = _SOFA.get(type_, type_)

        # start and stop
        if bits[3] == '.': raise IOError('invalid annotation line: start must be defined [line: {0}]'.format(self._lineno))
        if bits[4] == '.': raise IOError('invalid annotation line: stop must be defined [line: {0}]'.format(self._lineno))
        try:
            start = int(bits[3]) - 1
            stop = int(bits[4]) - 1
        except ValueError: raise IOError('invalid feature positions [line: {0}]'.format(self._lineno))
            # positions are checked later (after Is_circular can be detected)

        # score
        if bits[5] == '.': score = None
        else:
            try: score = float(bits[5])
            except ValueError: raise IOError('invalid feature score [line: {0}]'.format(self._lineno))

        # strand
        strand = bits[6]
        if strand not in '+-.?': raise IOError('invalid feature strand [line: {0}]'.format(self._lineno))
        if strand == '.': strand = None

        # phase
        phase = bits[7]
        if phase not in '.012': raise IOError('invalid feature phase [line: {0}]'.format(self._lineno))
        if type_ == 'CDS':
            if phase == '.':
                if self._strict: raise IOError('invalid feature: phase must be defined for CDS [line: {0}]'.format(self._lineno))
            else:
                phase = int(phase)
        else:
            if phase != '.' and type_ not in ['start_codon', 'stop_codon']:
                raise IOError('invalid feature: phase must only be defined for CDS [line: {0}]'.format(self._lineno))
        if phase == '.': phase = None

        # attributes
        ID = None
        Name = None
        Alias = None
        Parent = None
        Target = None
        Gap = None
        Note = None
        Dbxref = None
        Ontology_term = None
        Derives_from = None
        Is_circular = None
        attributes = {}
        attrs = bits[8].split(';')
        for attr in attrs:
            if attr.count('=') != 1: raise IOError('invalid attribute: {0} [line: {1}]'.format(attr, self._lineno))
            key, value = attr.split('=')
            if key[0].isupper():
                if key == 'ID':
                    if ID is not None: raise IOError('invalid attribute: {0} (ID defined more than once) [line: {1}]'.format(attr, self._lineno))
                    ID = self._process_text(value)
                elif key == 'Name':
                    if Name is not None: raise IOError('invalid attribute: {0} (Name defined more than once) [line: {1}]'.format(attr, self._lineno))
                    Name = self._process_text(value)
                elif key == 'Alias':
                    if Alias is not None: raise IOError('invalid attribute: {0} (Alias defined more than once) [line: {1}]'.format(attr, self._lineno))
                    Alias = tuple(map(self._process_text, value.split(',')))
                elif key == 'Parent':
                    if Parent is not None: raise IOError('invalid attribute: {0} (Parent defined more than once) [line: {1}]'.format(attr, self._lineno))
                    Parent = tuple(map(self._process_text, value.split(',')))
                elif key == 'Target':
                    if Target is not None: raise IOError('invalid attribute: {0} (Target defined more than once) [line: {1}]'.format(attr, self._lineno))
                    bits = value.split()
                    if len(bits) < 3 or len(bits) > 4:
                        raise IOError('invalid attribute: {0} [line: {1}]'.format(attr, self._lineno))
                    target_id = self._process_text(bits[0])
                    try:
                        target_start = int(bits[1]) - 1
                        target_end = int(bits[2]) - 1
                    except ValueError: raise IOError('invalid attribute: {0} (invalid bounds) [line: {1}]'.format(attr, self._lineno))
                    if target_start < 0 or target_end < target_start:
                        raise IOError('invalid attribute: {0} (invalid bounds) [line: {1}]'.format(attr, self._lineno))
                    if len(bits) == 3:
                        target_strand = '.'
                    else:
                        if bits[3] not in '+-':  raise IOError('invalid attribute: {0} [line: {1}]'.format(attr, self._lineno))
                        target_strand = bits[3]
                    Target = (target_id, target_start, target_end, target_strand)
                elif key == 'Gap':
                    if Gap is not None: raise IOError('invalid attribute: {0} (Gap defined more than once) [line: {1}]'.format(attr, self._lineno))
                    bits = value.split()
                    for bit in bits:
                        if re.match('[MIDFR]\d+$', bit) is None:
                            raise IOError('invalid attribute: {0} (invalid CIGAR string) [line: {1}]'.format(attr, self._lineno))
                    Gap = value
                elif key == 'Derives_from':
                    if Derives_from is not None: raise IOError('invalid attribute: {0} (Derives_from defined more than once) [line: {1}]'.format(attr, self._lineno))
                    Derives_from = self._process_text(value)
                elif key == 'Note':
                    if Note is not None: raise IOError('invalid attribute: {0} (Note defined more than once) [line: {1}]'.format(attr, self._lineno))
                    Note = tuple(map(self._process_text, value.split(',')))
                elif key == 'Dbxref':
                    if Dbxref is not None: raise IOError('invalid attribute: {0} (Dbxref defined more than once) [line: {1}]'.format(attr, self._lineno))
                    Dbxref = []
                    for v in value.split(','):
                        v = self._process_text(v)
                        if v.count(':') != 1: raise IOError('invalid attribute: {0} [line: {1}]'.format(attr, self._lineno))
                        Dbxref.append(v)
                elif key == 'Ontology_term':
                    if Ontology_term is not None: raise IOError('invalid attribute: {0} (Ontology_term defined more than once) [line: {1}]'.format(attr, self._lineno))
                    Ontology_term = []
                    for v in value.split(','):
                        v = self._process_text(v)
                        if v.count(':') != 1: raise IOError('invalid attribute: {0} [line: {1}]'.format(attr, self._lineno))
                        Ontology_term.append(v)
                elif key == 'Is_circular':
                    if Is_circular is not None: raise IOError('invalid attribute: {0} (Is_circular defined more than once) [line: {1}]'.format(attr, self._lineno))
                    if value.lower() == 'true': Is_circular = True
                    elif value.lower() == 'false': Is_circular = False
                    else: raise IOError('invalid attribute: {0} [line: {1}]'.format(key, self._lineno))
                else:
                    raise IOError('invalid attribute name: {0} [line: {1}]'.format(key, self._lineno))
            else:
                if key in attributes: raise IOError('invalid attribute: {0} ({1} defined more than once) [line: {1}]'.format(attr, key, self._lineno))
                attributes[key] = self._process_text(value)

        # default value Is_circular
        if Is_circular is None: Is_circular = False

        # check bounds
        if stop < start: raise IOError('invalid feature positions [line: {0}]'.format(self._lineno))
        if seqid in self._regions:
            if start < self._regions[seqid][0]: raise IOError('invalid feature positions [line: {0}]'.format(self._lineno))
            if stop > self._regions[seqid][1] and not Is_circular: raise IOError('invalid feature positions [line: {0}]'.format(self._lineno))

        if ID is not None:
            # check if new segment of an open feature
            if ID in self._open_features:
                f = self._open_features[ID]
                if seqid is not None and seqid != f._seqid: raise IOError('invalid feature: mismatch of `seqid` between feature segments [line: {0}]'.format(self._lineno))
                if type_ is not None and type_ != f._type: raise IOError('invalid feature: mismatch of `type` between feature segments [line: {0}]'.format(self._lineno))
                if score is not None and score != f._score: raise IOError('invalid feature: mismatch of `score` between feature segments [line: {0}]'.format(self._lineno))
                if strand is not None and strand != f._strand: raise IOError('invalid feature: mismatch of `strand` between feature segments [line: {0}]'.format(self._lineno))
                for k in attributes:
                    if k not in f.attributes or attributes[k] != f._attributes[k]:
                        if k in ['exon_number', 'exon_id']:
                            pass # ignore exon_number qualifier at the level of segment
                        else:
                            raise IOError('invalid feature: mismatch of `{0}` between feature segments [line: {1}]'.format(k, self._lineno))
                f._add_segment(start, stop, phase)
                return

            # check that ID not duplicated
            if ID in self._IDs: raise IOError('invalid feature: duplicated ID: {0} [line: {1}]'.format(ID, self._lineno))
            self._IDs.add(ID)

        # create new feature
        f = Gff3Feature._make(seqid=seqid, source=source, type_=type_,
            score=score, strand=strand, ID=ID, Name=Name, Alias=Alias,
            Parent=Parent, Target=Target, Gap=Gap,
            Derives_from=Derives_from, Note=Note, Dbxref=Dbxref,
            Ontology_term=Ontology_term, Is_circular=Is_circular,
            **attributes)
        f._add_segment(start, stop, phase)

        # connect to parents
        parents = []
        if Parent is not None:
            for p in Parent:
                if p not in self._open_features: raise IOError('invalid feature: invalid parent ID: {0} (non-existing or closed feature) [line: {1}]'.format(p, self._lineno))
                parents.append(self._open_features[p])
        derives = []
        if Derives_from is not None:
            for d in Derives_from:
                if d not in self._open_features: raise IOError('invalid feature: invalid `Derives_from` ID: {0} (non-existing or closed feature) [line: {1}]'.format(d, self._lineno))
                derives.append(self._open_features[d])
        f._connect(parents, derives)
        ultimate = []
        for parent in parents: parent._ultimate_parents(ultimate)
        for p in ultimate: p._all_parts.append(f)

        # add feature
        if ID is None: f._close()
        else: self._open_features[ID] = f
        self._num_tot += 1
        if len(parents) == 0:
            self._num_top += 1
            if seqid not in self._seqid:
                self._seqid.append(seqid)
                self._mapping[seqid] = []
            self._mapping[seqid].append(f)

    def _get_sequences(self): # import fasta-formatted sequences
        fp = _eggwrapper.FastaParser()
        try:
            fp.open_file(self._f.name, alphabets.DNA._obj, self._pos)
            self._f.seek(0, os.SEEK_END)
        except AttributeError:
            self._f.seek(self._pos)
            fp.set_string(self._f.read(), alphabets.DNA._obj)
        obj = _eggwrapper.DataHolder()
        fp.read_all(False, obj)
        self._sequences = _interface.Container._create_from_data_holder(obj, alphabets.DNA)

    @property
    def version(self):
        """GFF version."""
        return self._version

    @property
    def regions(self):
        """Dictionary with all sequence regions defined by directives."""
        return self._regions

    @property
    def feature_ontology(self):
        """Content of all feature-ontology directives."""
        return self._feature_ontology

    @property
    def attribute_ontology(self):
        """Content of all attribute-ontology directives."""
        return self._attribute_ontology

    @property
    def source_ontology(self):
        """Content of all source-ontology directives."""
        return self._source_ontology

    @property
    def species(self):
        """Species (if specified by a directive)."""
        return self._species

    @property
    def genome_build(self):
        """Genome build if specified by a directive."""
        return self._genome_build

    @property
    def sequences(self):
        """
        If specified, sequences as a :class:`.Container`. By default,
        this value is ``None``.
        """
        return self._sequences

    @property
    def num_seqid(self):
        """Number of seqid of features present in instance."""
        return len(self._seqid)

    @property
    def list_seqid(self):
        """List of seqid."""
        return self._seqid

    @property
    def num_top_features(self):
        """Number of features without parents."""
        return self._num_top

    @property
    def num_tot_features(self):
        """Total number of features."""
        return self._num_tot

    def __iter__(self):
        return Iterator(self, None, None, None, False)

    def iter_features(self, seqid=None, start=None, end=None, all_features=False):
        """
        Iterate over features.

        :param seqid: only iterates overs features of this contig (by
            default, consider all contigs). If the specified seqid is not
            present in the GFF3 file, iteration is empty (without error).
        :param start: start iteration at this position. By default,
            start with the first position.
        :param end: stop iteration at this position (does not include
            features whose end position is larger than this value). By
            default, process all features.
        :param all_features: whether iteration should also include
            features that are part of another (by default, only include
            features that don't have a parent). A given feature may be
            repeated if it has several unconnected parents.

        .. note::
            If *start* and/or *end* are specified, *seqid* is required.
        """
        return Iterator(self, seqid, start, end, all_features)

class Iterator(object):
    def __init__(self, obj, seqid, start, end, all_features):

        # check that bounds are used only for one seqid
        if (start is not None or end is not None) and seqid is None:
            raise ValueError('if start/end are specified, seqid must be specified')
        if start is None: start = 0
        if end is None: end = _eggwrapper.MAX
        self._obj = obj
        self._end = end

        # get list of seqid to process
        if seqid is None: self._seqid = obj._seqid
        elif seqid not in obj._seqid: self._seqid = []
        else: self._seqid = [seqid]
        self._idx_seq = 0
        if len(self._seqid) == 0:
            self._next = self._nextD
            return

        # set to the start position
        if start > 0:
            if all_features:
                for self._idx_top in range(len(obj._mapping[self._seqid[0]])):
                    for self._idx_part in range(len(obj._mapping[self._seqid[0]][self._idx_top].all_parts)):
                        if obj._mapping[self._seqid[0]][self._idx_top].all_parts[self._idx_part].start >= start: break
                    else: continue # no part has been found: check next top feature
                    break # a part has been found
                else: # no features found after the start position
                    self._next = self._nextD
            else:
                for self._idx_top in range(len(obj._mapping[self._seqid[0]])):
                    if obj._mapping[self._seqid[0]][self._idx_top].start >= start: break
                else: # no features found after the start position
                    self._next = self._nextD
        else:
            self._idx_top = 0
            self._idx_part = 0

        if all_features:
            self._next = self._nextAF
        else:
            self._next = self._nextTF

    def __iter__(self):
        return self

    def __next__(self):
        return self._next()

    next = __next__

    def _nextD(self):
        raise StopIteration

    def _nextAF(self):
        if self._idx_seq == len(self._seqid): raise StopIteration
        feats = self._obj._mapping[self._seqid[self._idx_seq]]
        if self._idx_top == len(feats):
            self._idx_seq += 1
            self._idx_top = 0
            return self._nextAF()
        if feats[self._idx_top].start > self._end: raise StopIteration
        if self._idx_part == len(feats[self._idx_top].all_parts):
            self._idx_top += 1
            self._idx_part = 0
            return self._nextAF()
        self._idx_part += 1
        if feats[self._idx_top].all_parts[self._idx_part-1].end <= self._end:
            return feats[self._idx_top].all_parts[self._idx_part-1]
        return self._nextAF()

    def _nextTF(self):
        if self._idx_seq == len(self._seqid): raise StopIteration
        feats = self._obj._mapping[self._seqid[self._idx_seq]]
        if self._idx_top == len(feats):
            self._idx_seq += 1
            self._idx_top = 0
            return self._nextTF()
        if feats[self._idx_top].end > self._end: raise StopIteration # necessarily finished because there must be only one seqid
        self._idx_top += 1
        return feats[self._idx_top-1]

class Gff3Feature(object):
    """
    Provide access to data of a feature. This class cannot be instanciated
    by the user.
    """
    def __init__(self):
        raise ValueError('Gff3Feature instances cannot be created directly')

    @classmethod
    def _make(cls, seqid, source, type_, score, strand, ID, Name,
                 Alias, Parent, Target, Gap, Derives_from, Note,
                 Dbxref, Ontology_term, Is_circular, **attrs):
        obj = Gff3Feature.__new__(cls)
        obj._seqid = seqid
        obj._type = type_
        obj._score = score
        obj._source = source
        obj._strand = strand
        obj._ID = ID
        obj._attributes = {'ID': ID, 'Name': Name, 'Alias': Alias,
            'Parent': Parent, 'Target': Target, 'Gap': Gap,
            'Derives_from': Derives_from, 'Note': Note,
            'Dbxref': Dbxref, 'Ontology_term': Ontology_term,
            'Is_circular': Is_circular}
        obj._attributes.update(attrs)
        obj._start = None
        obj._end = None
        obj._segments = []
        obj._descendants = []
        obj._all_parts = [obj]
        obj._derivers = []
        return obj

    @property
    def seqid(self):
        """Seqid on which this feature is located."""
        return self._seqid

    @property
    def type(self):
        """Feature type."""
        return self._type

    @property
    def score(self):
        """Feature score."""
        return self._score

    @property
    def source(self):
        """Feature source."""
        return self._source

    @property
    def strand(self):
        """Strand on which the feature is located."""
        return self._strand

    @property
    def ID(self):
        """Feature identifier."""
        return self._ID

    @property
    def attributes(self):
        """
        Dictionary of attributes attached to the feature.
        The list of attributes is:

        * ID
        * Name
        * Alias
        * Parent
        * Target
        * Gap
        * Derives_from
        * Note
        * Dbxref
        * Ontology_term
        * Is_circular
        * Other attributes as defined in the GFF3 file.
        """
        return self._attributes

    @property
    def start(self):
        """Feature start position."""
        return self._start

    @property
    def end(self):
        """Feature end position."""
        return self._end

    @property
    def segments(self):
        """List of segments of the feature."""
        return self._segments

    @property
    def parents(self):
        """List of features parent to this one."""
        return self._parents

    @property
    def descendants(self):
        """List of features descending from this one."""
        return self._descendants

    @property
    def all_parts(self):
        """All parts of this feature."""
        return self._all_parts

    @property
    def derivers(self):
        """List of features deriving from this one."""
        return self._derivers

    def _add_segment(self, start, end, phase):
        # Add a segment to the feature. Assume that start <= end. The
        # segments can be loaded in any order. Segments are not allowed to
        # overlap. They are sorted when _close() is called.
        self._segments.append((start, end, phase))

    def _connect(self, parents, derives):
        # Connect this feature to parents or other features from which it
        # derives.
        self._parents = parents
        self._derives_from = derives
        for i in self._parents: i._descendants.append(self)
        for i in self._derives_from: i._derivers.append(self)

    def _ultimate_parents(self, dest):
        # Return the list of ultimate parents.
        if len(self._parents) == 0: dest.append(self)
        else:
            for p in self._parents: p._ultimate_parents(dest)

    def _close(self):
        # Terminate loading of segments in the instance.
        if len(self._segments) == 0: raise ValueError('feature without segments')
        self._segments.sort(key=operator.itemgetter(0))
        for i in range(1, len(self._segments)):
            if self._segments[i-1][1] >= self._segments[i][0]: raise ValueError('feature segments are overlapping')
        self._start = self._segments[0][0]
        self._end = self._segments[-1][1]



"""
    Copyright 2015-2021 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import eggwrapper as _eggwrapper
from .. import _interface, alphabets
from .. import _site

alphabet_genotypes = alphabets.Alphabet('range', (0, None), (-1, 0),
    case_insensitive=False, name='genotypes')

def haplotypes_from_align(aln, max_missing=0.0, impute_threshold=0, struct=None, dest=None, multiple=False):
    """
    Identification of haplotypes from an alignment.
    Identify haplotypes an :class:`.Align` instance and return data as a
    single :class:`.Site` instance containing one sample for each sample
    of the original data. Alleles in the returned site are representing
    all identified haplotypes (or missing data when the haplotypes could
    not be derived).

    .. note::
        There must be at least one site with at least two alleles (overall,
        including the outgroup), otherwise the produced site only contains
        missing data.

    :param aln: an :class:`.Align` instance.

    :param impute_threshold: by default, all samples with a least one
        occurrence of missing data will be treated as missing data. If
        this argument is more than 0, the provided value will be used as
        maximum number of missing data. All samples with as many or less
        missing data will be processed to determine which extant of
        haplotype they might belong (to which they are identical save
        for missing data). If there is only one such haplotype, the
        corresponding samples will be treated as a repetition of this
        haplotype. This option will never allow detecting new
        haplotypes. Only small values of this option make sense.

    :param struct: a :class:`.Structure` instance defining the samples
        to process. The population and cluster structures are not used.
        If the ploidy is larger than 1, the individuals are used, and
        sites are assumed to be phased.

    :param max_missing: maximum relative proportion of missing data in the ingroup
        to process a site.

        .. note::
            Here, *max_missing* is a relative proportion to allow
            using the same value for different alignments that might
            not have the same number of samples (to avoid reevaluating
            *max_missing* if the user wants the same maximum rate
            of missing data). In other functions, *max_missing* is
            the maximum *number* of missing and is an integer.

    :param dest: a :class:`.Site` instance that will be reset and in
        which data will be placed. If specified, this function returns
        nothing. By default, the function returns a new :class:`.Site`
        instance.

    :param multiple: allow sites with more than two alleles in the
        ingroup.

    :return: A :class:`.Site` instance, (if *dest* is ``None``) or ``None``
        (otherwise).

    .. note::
        If an empty list of sites is provided and no structure is provided,
        a :exc:`ValueError` is raised. To support smoothly possibly empty lists,
        the used is prompted to provide a structure.
    """
    if not isinstance(aln, _interface.Align): raise TypeError('expect an Align instance')
    return _haplotypes(aln, max_missing, impute_threshold, struct, dest, multiple)

def haplotypes_from_sites(sites, impute_threshold=0, struct=None, dest=None, multiple=False):
    """
    Identification of haplotypes from a list of sites.
    Similar to :meth:`haplotypes_from_align` but takes a list of :class:`.Site`
    instances. No option *max_missing* is available; all sites are always considered.
    """
    if not isinstance(sites, list): raise TypeError('expect a list of Site instances')
    if set(map(type, sites)) != set([_site.Site]): raise TypeError('expect a list of Site instances')
    return _haplotypes(sites, 1.0, impute_threshold, struct, dest, multiple)

def _haplotypes(sites, max_missing=0.0, impute_threshold=0, struct=None, dest=None, multiple=False):

    # check arguments / init
    if max_missing < 0.0 or max_missing > 1.0: raise ValueError('max_missing argument out of range')
    if impute_threshold < 0: raise ValueError('invalid value for impute_threshold argument')
    obj = _eggwrapper.Haplotypes()
    frq = _eggwrapper.FreqHolder()
    if struct is None:
        if isinstance(sites, _interface.Align):
            ni = sites._obj.get_nsam()
        else:
            if len(sites) == 0: raise ValueError('cannot process an empty list of sites')
            ni = sites[0]._obj.get_ns()
        struct_obj = _eggwrapper.StructureHolder()
        struct_obj.mk_dummy_structure(ni, 1)
        struct2_obj = struct_obj # used only with haplotypes_from_align
        pl = 1
        no = 0
    else:
        struct_obj = struct._obj
        struct2_obj = struct.make_sorted_auxiliary()._obj # used only with haplotypes_from_align
        pl = struct_obj.get_ploidy()
        ni = struct_obj.num_indiv_ingroup()
        no = struct_obj.num_indiv_outgroup()
    obj.setup(struct_obj)
    impute_idx = []
    min_exploitable = ni * pl - int(max_missing * pl * ni)

    # pass sites if they come from an alignment
    genosite = _eggwrapper.Genotypes()
    if isinstance(sites, _interface.Align):
        if struct_obj.get_req() > sites._obj.get_nsam(): raise ValueError('structure object does not match alignment')
        site = _eggwrapper.SiteHolder()
        for i in range(sites._obj.get_nsit_all()):
            site.reset()
            if site.process_align(sites._obj, i, struct_obj) < min_exploitable: # process_align returns number of valid ingroup samples
                continue
            frq.setup_structure(struct2_obj)
            frq.process_site(site)
            genosite.process(site, struct2_obj)
            nall = frq.frq_ingroup().num_alleles_eff()
            if frq.num_alleles() > 1 and (multiple == True or nall < 3):
                obj.load(genosite)
                if impute_threshold > 0: impute_idx.append(i)

        if obj.n_sites() == 0:
            if dest is None:
                return _site.site_from_list([-1 for i in range(ni+no) for j in range(pl)],
                                                alphabet = alphabet_genotypes)
            else:
                dest.process_list([-1 for i in range(ni+no) for j in range(pl)],
                                        alphabet = alphabet_genotypes, reset=True)
                return None

    # process list of Site
    else:
        if len(sites) > 0: ns = sites[0]._obj.get_ns()
        for i, site in enumerate(sites):
            if site._obj.get_ns() != ns: raise ValueError('number of samples must be consistent between sites')
            frq.setup_structure(struct_obj)
            frq.process_site(site._obj)
            if frq.frq_ingroup().nseff()+frq.frq_outgroup().nseff() < min_exploitable:
                continue
            genosite.process(site._obj, struct_obj)
            nall = frq.frq_ingroup().num_alleles_eff()
            if frq.num_alleles() > 1 and (multiple == True or nall < 3):
                obj.load(genosite)
                if impute_threshold > 0: impute_idx.append(i)

    # complete haplotype detection
    obj.cp_haplotypes()

    # impute if requested
    if impute_threshold > 0:
        obj.prepare_impute(impute_threshold)
        for i in impute_idx:
            if isinstance(sites, _interface.Align):
                site.reset(pl)
                site.process_align(sites._obj, i, _eggwrapper.MAX, True)
                obj.resolve(site)
            else:
                obj.resolve(sites[i]._obj)
        obj.impute()

    # generate site
    if dest is None:
        dest = _site.Site()
        dest._alphabet = alphabet_genotypes
        obj.get_site(dest._obj)
        return dest
    else:
        dest._alphabet = alphabet_genotypes
        obj.get_site(dest._obj)



"""
    Copyright 2009-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import subprocess, re, os
from .. import _interface, alphabets
from ..io import _fasta
from . import _utils

class _Muscle(_utils._App):

    @_utils._protect_run
    def _check_path(self, path, cfg):

        # test the "help" option to ensure that software exists
        cmd = (path, '-version')
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
            stdout, stderr = p.communicate()
            if len(stderr): return stderr
        except OSError as e:
            return e.strerror

        if re.match('MUSCLE v3\.8\.\d+', stdout):
            cfg['version'] = 3
        elif re.match('muscle 5\.\d+\.', stdout):
            cfg['version'] = 5
        else:
            return 'MUSCLE: invalid version'
        return None

_app = _Muscle(key='muscle', default='muscle')
_app.add_option(name='version', f_convert=int, default=None, f_set_missing=lambda: 5, f_load_missing=lambda: 5)
_utils.paths._add(_app)

# dynamic caller who redirected call to the correct version of the wrapper
def muscle(*args, **kwargs):
    """ muscle(...)
    Perform multiple alignment using `Muscle <http://www.drive5.com/muscle/>`_.

    Depending of the version of MUSCLE detected at configuration, the
    call will be forwarded to either :func:`.wrappers.muscle3` or
    :func:`.wrappers.muscle5`.

    .. versionchanged:: 3.2.0

        Dynamically use the new :func:`!muscle5` method if MUSCLE
        version 5 is present.
    """

    if _app.config['version'] == 5: return muscle5(*args, **kwargs)
    elif _app.config['version'] == 3: return muscle3(*args, **kwargs)
    else: raise RuntimeError('muscle program not available -- please configure path')

##### muscle 3 stuff ###################################################

_args = {
    # these ones cannot passed as is to muscle
    'diags':           (bool,  None),
    'diags1':          (bool,  None),
    'diags2':          (bool,  None),
    'anchors':         (bool,  None),
    'dimer':           (bool,  None),
    'brenner':         (bool,  None),
    'cluster':         (bool,  None),
    'teamgaps4':       (bool,  lambda x: x is True),
    'nt_profile':      (str,   ['spn']),
    'aa_profile':      (str,   ['le', 'sp', 'sv']),

    # these ones should be able to be passed as is
    'anchorspacing':   (int,   lambda x: x>0),
    'center':          (float, lambda x: x<0),
    'cluster1':        (str,   ['upgma', 'upgmb', 'neighborjoining']),
    'diagbreak':       (int,   lambda x: x>0),
    'diaglength':      (int,   lambda x: x>0),
    'diagmargin':      (int,   lambda x: x>0),
    'diagmargin':      (int,   lambda x: x>0),
    'distance1':       (str,   ['kmer6_6', 'kmer20_3', 'kmer20_4', 'kbit20_3', 'kmer4_6']),
    'distance2':       (str,   ['pctidkimura', 'pctidlog']),
    'gapopen':         (float, lambda x: x<0),
    'hydro':           (int,   lambda x: x>0),
    'hydrofactor':     (float, lambda x: x>0),
    'maxiters':        (int,   lambda x: x>0),
    'maxtrees':        (int,   lambda x: x>0),
    'minbestcolscore': (float, lambda x: x>=0),
    'minsmoothscore':  (float, lambda x: x>=0),
    'objscore':        (str,   ['sp', 'ps', 'dp', 'xp', 'spf', 'spm']),
    'refinewindow':    (int,   lambda x: x>0),
    'root1':           (str,   ['pseudo', 'midlongestspan', 'minavgleafdist']),
    'seqtype':         (str,   ['protein', 'dna', 'auto']),
    'smoothscoreceil': (float, lambda x: x>0),
    'SUEFF':           (float, lambda x: x>=0.0 and x<=1.0),
    'weight1':         (str,   ['none', 'henikoff', 'henikoffpb', 'gsc', 'clustalw', 'threeway']),
    'weight2':         (str,   ['none', 'henikoff', 'henikoffpb', 'gsc', 'clustalw', 'threeway'])
}

@_utils._protect_run
def muscle3(source, ref=None, verbose=False, **kwargs):

    """ muscle(source, ref=None, verbose=False, \
                **kwargs)

    Perform multiple alignment using `Muscle <http://www.drive5.com/muscle/>`_.

    This wrapper is designed to run version 3 of MUSCLE. To use MUSCLE
    version 5, configure the application path using the
    ``egglib-config apps`` command.

    MUSCLE's default options tend to produce high-quality alignments but
    may be long to run on large data sets. Muscle's author recommends
    using the option ``maxiters=2`` for large data sets, and, for fast
    alignment (in particular of closely related sequences):
    ``maxiters=1 diags=True aa_profile='sv' distance1='kbit20_3'`` (for
    amino acid sequences) and ``maxiters=1 diags=True`` (for nucleotide
    sequences).

    :param source: a :class:`.Container` or :class:`.Align` containing
        sequences to align. If an :class:`!Align` is provided, sequences
        are assumed to be already aligned and alignment will be refined
        (using the ``-refine`` option of Muscle), unless an alignment is
        also provided as *ref*. In the latter case,  the two alignments
        are preserved (their columns are left unchanged), and they are
        aligned with respect to each other.

    :param ref: an :class:`!Align` instance providing an alignement that
        should be aligned with respect to the alignment provided as
        *source*. If *ref* is provided, it is required both *source* and
        *ref* are :class:`!Align` instances.

    :param verbose: display Muscle's console output.

    :param kwargs: other keyword arguments are passed to Muscle. The
        available options are listed below:

        .. include:: muscle_arguments.txt

        For a description of options, see the `Muscle manual <http://www.drive5.com/muscle/manual/options.html>`_.
        Most of Muscle's options are available. Note that function takes
        no flag option, and Muscle's flag options are passed as boolean
        keyword arguments (except options relative to the amino acid or
        nucleotide profile score options, that are passed as string as
        ``aa_profile`` and ``nt_profile``, respectively. The order of
        options is preserved.

    :return: An :class:`.Align` containing aligned sequences.

    .. versionchanged:: 3.0.0

        Added support for most options.

    .. versionchanged:: 3.2.0

        Renamed as :func:`!muscle3`. Available as :func:`!muscle` if
        MUSCLE version 3 is available
    """

    # check that program is available
    if _app.config['version'] != 3:
        raise RuntimeError('muscle version 3 not available -- please configure path')
    command_line = [_app.get_path(), '-out', 'o']
    if not verbose: command_line.append('-quiet')

    # mapping for sample reference
    mapping = {}

    # write source
    if not isinstance(source, (_interface.Container, _interface.Align)): raise TypeError('invalid type for `source` argument')
    _utils._write(source, 'source', mapping)

    # then save ref
    if ref != None:
        if not isinstance(ref, _interface.Align): raise TypeError('`ref` must be an Align')
        if not isinstance(source, _interface.Align): raise TypeError('if `ref` is provided, `source` must be an Align')
        if ref._alphabet not in [alphabets.DNA, alphabets.protein]: raise ValueError('alphabet must be DNA or protein')
        _utils._write(ref, 'ref', mapping)
        command_line.extend(['-profile', '-in1', 'ref', '-in2', 'source'])

    elif isinstance(source, _interface.Align):
        command_line.extend(['-in', 'source', '-refine'])

    else: # the default case
        command_line.extend(['-in', 'source'])

    # check data type
    if ref != None and ref.alphabet != source.alphabet: raise ValueError('`source` and `ref` must have the same alphabet')
    if 'seqtype' not in kwargs:
        if source.alphabet == alphabets.DNA: kwargs['seqtype'] = 'dna'
        elif source.alphabet == alphabets.protein: kwargs['seqtype'] = 'protein'
        else: raise ValueError('unsupported alphabet')
    else:
        if kwargs['seqtype'] == 'dna' and source.alphabet != alphabets.DNA: raise ValueError('source does not have a proper alphabet for seqtype `{0}`'.format(kwargs['seqtype']))
        if kwargs['seqtype'] == 'protein' and source.alphabet != alphabets.protein: raise ValueError('source does not have a proper alphabet for seqtype `{0}`'.format(kwargs['seqtype']))

    # process kwargs
    for opt, value in kwargs.items():
        if opt not in _args: raise ValueError('invalid option: {0}'.format(opt))
        type_, test = _args[opt]
        try: value = type_(value)
        except (TypeError, ValueError): raise ValueError('invalid value for argument {0}'.format(opt))
        if type_ == str:
            if value not in test: raise ValueError('invalid value for argument {0}'.format(opt))
        elif test is not None and not test(value): raise ValueError('invalid value for argument {0}'.format(opt))

        if opt == 'diags':
            if value == True: command_line.append('-diags')
        elif opt == 'diags1':
            if value == True: command_line.append('-diags1')
        elif opt == 'diags2':
            if value == True: command_line.append('-diags2')
        elif opt == 'anchors':
            if value == True: command_line.append('-anchors')
            else: command_line.append('-noanchors')
        elif opt == 'dimer':
            if value == True: command_line.append('-dimer')
        elif opt == 'brenner':
            if value == True: command_line.append('-brenner')
        elif opt == 'cluster':
            if value == True: command_line.append('-cluster')
        elif opt == 'teamgaps4':
            if value == True: command_line.append('-teamgaps4')
        elif opt == 'nt_profile':
            command_line.append('-' + value)
        elif opt == 'aa_profile':
            command_line.append('-' + value)
        else:
            command_line.extend(['-' + opt, str(value)])

    # run the program
    p = subprocess.Popen(command_line, stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=(None if verbose else subprocess.PIPE), universal_newlines=True)
    stdout, stderr = p.communicate('Y\n')

    # check error
    if not os.path.isfile('o'):
        if stderr is not None:
            lines = stderr.splitlines()
            if len(lines) and 'ERROR' in lines[-1]:
                raise RuntimeError('error while running muscle: {0}'.format(lines[-1].strip()))
        raise RuntimeError('unknown error while running muscle (try verbose mode)')

    # get alignment
    aln = _fasta.from_fasta('o', source.alphabet, labels=False, cls=_interface.Align)

    # set proper names and groups
    for sam in aln:
        ref = mapping[sam.name]
        sam.name = ref.name
        sam.labels = ref.labels

    # return
    return aln

##### muscle 5 stuff ###################################################

@_utils._protect_run
def muscle5(source, super5=False, perm='none', perturb=0, consiters=2,
                        refineiters=100, threads=None, verbose=False):

    """ muscle5(source, super5=False, perm='none', perturb=0, consiters=2, \
                        refineiters=100, threads=None, verbose=False)

    Perform multiple alignment using `Muscle <http://www.drive5.com/muscle/>`_ version 5.

    :param source: a :class:`.Container` instance contain sequences to
        align. :class:`.Align` is supported but will be treated as if it
        was a :class:`!Container`. The alphabet must be DNA or protein.

    :param super5: use the `Super5 algorithm <https://drive5.com/muscle5/manual/super5_algo.html>`_
        (recommended for datasets of more than a few hundred sequences).

    :param perm: guide tree permutation mode. Available values are
        ``none``, ``abc``, ``acb``, and ``bca``. More information
        `here <https://drive5.com/muscle5/manual/guide_tree_permutations.html>`_.

    :param perturb: if different of 0, the value is a random number
        generator seed used to perform `hidden Markov model perturbations <https://drive5.com/muscle5/manual/hmm_perturbations.html>`_.

    :param consiter: number of consistency iterations.

    :param refineiters: number of refinement iterations.

    :param threads: number of threads. By default, let MUSCLE pick the
        value.

    :param verbose: show MUSCLE's output in stdout.

    .. note::
        The *ensemble fasta* features of muscle5 are currently not
        available through this wrapper.
    
    :return: An :class:`.Align` containing aligned sequences.

    .. versionadded:: 3.2.0
    """

    # check version (implies presence of program)
    if _app.config['version'] != 5:
        raise RuntimeError('muscle version 5 not available -- please configure path')

    # initialize command line
    command_line = [_app.get_path(),
                    '-super5' if super5 else '-align',
                    'i', '-output', 'o']

    # check input object
    if not isinstance(source, (_interface.Container, _interface.Align)):
        raise TypeError('`source\' must be a Container')
    if source._alphabet == alphabets.DNA: pass
    elif source._alphabet == alphabets.protein: pass
    else: raise ValueError('alphabet must be DNA or protein')
    _utils._write(source, 'i', mapping:={})

    # permutation mode
    if perm != 'none':
        if perm not in muscle5_perm_values:
            raise ValueError('invalid value for `perm\'')
        command_line.extend(['-perm', perm])

    # HMM perturbations
    if not isinstance(perturb, int): raise TypeError('`perturb\' must be an integer')
    if perturb != 0:
        if perturb < 0: raise ValueError('invalid value for `perturb\'')
        command_line.extend(['-perturb', str(perturb)])

    # consiters / refineiters
    if not isinstance(consiters, int): raise TypeError('`consiters\' must be an integer')
    if not isinstance(refineiters, int): raise TypeError('`refineiters\' must be an integer')
    if consiters < 1: raise ValueError('`consiters\' must be strictly positive')
    if refineiters < 1: raise ValueError('`refineiters\' must be strictly positive')
    command_line.extend(['-consiters', str(consiters), '-refineiters', str(refineiters)])

    # threads
    if threads is not None:
        if not isinstance(threads, int): raise TypeError('`threads\' must be an integer')
        if threads < 1: raise ValueError('`threads\' must be strictly positive')
        command_line.extend(['-threads', str(threads)])

    # run the program
    p = subprocess.Popen(command_line, stdin=subprocess.PIPE,
                        stdout=(None if verbose else subprocess.PIPE),
                        stderr=(None if verbose else subprocess.PIPE), universal_newlines=True)
    stdout, stderr = p.communicate('Y\n')

    # check error
    if not os.path.isfile('o'):
        if stderr is not None:
            lines = stderr.splitlines()
            if len(lines) and 'ERROR' in lines[-1]:
                raise RuntimeError('error while running muscle: {0}'.format(lines[-1].strip()))
        raise RuntimeError('unknown error while running muscle (try verbose mode)')

    # get alignment
    aln = _fasta.from_fasta('o', source.alphabet, labels=False, cls=_interface.Align)

    # restore original names and groups
    for sam in aln:
        ref = mapping[sam.name]
        sam.name = ref.name
        sam.labels = ref.labels

    # return
    return aln

muscle5_perm_values = {'abc', 'acb', 'bca'}
